import posixpath
from typing import Any, Union, Tuple, List, Optional
from urllib.parse import unquote

from monolith_filemanager.adapters.base import Base
from monolith_filemanager.adapters.errors import S3ProcessesAdapterError
from monolith_filemanager.adapters.utils import monolith_globre_match
from monolith_filemanager.file.base import File, FilePath
from monolith_filemanager.s3storage import V1Engine


class S3ProcessesAdapter(Base):
    """
    This is a class for managing the interface of s3 file commands.
    """
    def __init__(self, file_path: Union[FilePath], caching: Optional[Any] = None) -> None:
        """
        The constructor for the S3ProcessesAdapter class.

        self.path has any trailing '/' stripped to account for root folder level paths e.g. s3://example-bucket/

        :param file_path: (str) path to the file concerned
        :param caching: (Optional[Any]) the CacheManager object to be used which is to be initialized before being passed through
        """
        super().__init__(file_path=file_path)
        self._engine: V1Engine = V1Engine()
        self._cache: Any = caching
        self._s3: bool = True
        self._strip_path_slash()

    def local_file_object(self) -> File:
        """
        Gets the reading and writing protocols based on the file type from the path.

        :return: Object containing read and write protocols for the type of file
        """
        file_object = self.file_types.get_file(file_path=self.path)
        return file_object(path=self.path)

    def read_file(self, **kwargs) -> Any:
        """
        Reads file from s3 storage.

        :return: data from file
        """
        if self.path.file_type == "pickle":
            try:
                from pickle_factory import base as pickle_factory
            except ImportError:
                raise S3ProcessesAdapterError(
                    message="you are trying to read a .pickle file without the legacy DPU pickle_factory plugin. "
                            "If you are not trying to pickle a DPU legacy object please use .sav instead of .pickle. "
                            "If you are trying to process DPU legacy objects please define the pickle_factory "
                            "in your PYTHONPATH")
            raw_data = self._engine.download_raw_data_file(storage_path=self.path.to_string())
            return pickle_factory.load(file=raw_data, memory=True)

        elif self.path.file_type == "parquet":
            # read_parquet can support s3 paths
            return self.local_file_object().read(**kwargs)

        else:
            if self._cache is None:
                self.raise_missing_cache_error(usage=f'read file \'{self.path}\'')
            self._cache.create_cache()
            stored_file_path = self._engine.download_data_file(storage_path=self.path.to_string(),
                                                               file_path=self._cache.cache_path)
            self.path = FilePath(stored_file_path)
            return self.local_file_object().read(**kwargs)

    def read_raw_file(self) -> Any:
        """
        Reads raw file from s3 storage.

        :return: (bytes) data from file
        """
        return self._engine.download_raw_data_file(storage_path=self.path.to_string())

    def custom_read_file(self, custom_read_function: Any) -> Any:
        """
        Passes a custom function and executes it in order to read the data.

        :param custom_read_function:
        :return:
        """
        if self._cache is None:
            self.raise_missing_cache_error(usage=f'custom read file \'{self.path}\'')
        self._cache.create_cache()
        stored_file_path = self._engine.download_data_file(storage_path=self.path,
                                                           file_path=self._cache.cache_path)
        self.path = stored_file_path
        return custom_read_function(self.path)

    def raise_missing_cache_error(self, usage: str):
        """
        Raise an error that the caching module is required but not provided.
        :param usage: (str) message about the task that was being attempted.
        :return: None
        """
        raise S3ProcessesAdapterError(
            f'You are trying to {usage} without providing a CacheManager object from the caching module. '
            f'Try the following:\n'
            '1) pip install the Monolith caching module;\n'
            '2) Create file objects passing an instance of `caching.CacheManager()` to `file_manager` method;\n'
            'For example:'
            '`file = general_filemanager.file_manager(file_path=file_path, caching=caching.CacheManager())`'
        )

    def write_file(self, data: Any, **kwargs) -> None:
        """
        Uploads data to s3 bucket.

        :param data: (Any) data to be uploaded to bucket
        :return: None
        """
        file_object = self.local_file_object()
        if file_object.supports_s3():
            return file_object.write(data, **kwargs)
        else:
            if self._cache is None:
                self.raise_missing_cache_error(usage=f'write file \'{self.path}\'')
            self._cache.create_cache()
            file_object.path = FilePath(self._cache.cache_path + file_object.path.split("/")[-1])
            file_object.write(data)
            self._engine.upload_data_from_file(storage_path=self.path.to_string(), file_path=file_object.path.to_string())

    def write_raw_file(self, data: Any) -> None:
        """
        Uploads raw data to s3 bucket directly.

        :param data: (Any) data to be uploaded to bucket
        :return: None
        """
        self._engine.upload_data(storage_path=self.path.to_string(), data=data)

    def delete_file(self, path: Optional[FilePath] = None) -> None:
        """
        Deletes file data from s3 bucket.

        :param path: (Optional[FilePath]) if None, defaults to self.path
        :return: None
        """
        if not path:
            path = self.path
        self._engine.delete(storage_path=path)

    def delete_folder(self, path: Optional[FilePath] = None) -> None:
        """
        Deletes "folder" and contents from s3 bucket.

        :param path: (Optional[FilePath]) if None, defaults to self.path
        :return: None
        """
        if not path:
            path = self.path
        bucket, folder_name, short_file_name = V1Engine._split_s3_path(storage_path=path)
        # ensure folder names being passed as s3 "Prefix" args are terminated with a "/"
        self._engine.delete_file(bucket_name=bucket, file_name=f"{folder_name}/")

    def write_stream(self, stream: Any) -> str:
        """
        Writes a stream of data to a cache and then uploads it to a s3 path.
        Checks that filename for path is not already taken by folder in the parent directory.
        If file with name already exists, name is incremented.

        :param stream: (Any) the stream to be written
        :return: (str) Name file saved as
        :raises: (S3ProcessesAdapterError) If file name already used by existing folder parent directory
        """
        if self._cache is None:
            self.raise_missing_cache_error(usage=f'write stream \'{self.path}\'')
        self._cache.create_cache()
        file_name = self.path.split("/")[-1]
        file_path = self._cache.cache_path + file_name
        parent_dir = FilePath("/".join(self.path.split("/")[:-1]) + "/")
        stream.save(file_path)
        files, dirs = self.ls(path=parent_dir)
        if self.check_name_taken(file_name, dirs):
            raise S3ProcessesAdapterError("New file name already taken by folder in this folder")
        if self.exists():
            self.increment_files()

        self._engine.upload_data_from_file(storage_path=self.path, file_path=file_path)
        file_name = self.path.split("/")[-1]
        return file_name

    def increment_files(self) -> None:
        """
        Increments file name integer suffix based on how many files with same filename prefix exist within S3 folder obj.
        Fills in gaps in contiguous numeric sequence first. Assigns final name to self.path.
        :return: None
        """
        count = 2
        prefix = "/".join(self.path.split("/")[:-1])
        ext = f".{self.path.split('.')[-1]}"
        file = ".".join(self.path.split("/")[-1].split(".")[:-1])
        new_file = FilePath(prefix + "/" + file + f" {count}{ext}")
        while self.exists(path=new_file):
            count += 1
            new_file = FilePath(prefix + "/" + file + f" {count}{ext}")
        self.path = new_file

    def create_directory_if_not_exists(self) -> str:
        """
        Creates a directory if it does not exist.
        :return: (str) returns name of newly created folder
        """
        return self._engine.create_folder(storage_path=self.path)

    def exists(self, path: Optional[FilePath] = None) -> bool:
        """
        Checks to see if the self.path exists.

        :return: (bool) True if exists, False if not
        """
        if path:
            return self._engine.exists(storage_path=path)
        else:
            return self._engine.exists(storage_path=self.path)

    def ls(self, path: Optional[FilePath] = None) -> Tuple[dict, List[str]]:
        """
        Lists all the sub directories and sub files belonging to the self.path.

        :param path: (Optional[str]) directory path if different from self.path
        :return: (Tuple[dict, List[str]]) sub files and sub directories
        """
        if path:
            return self._engine.ls(storage_path=path)
        else:
            return self._engine.ls(storage_path=self.path.to_string())

    def _extract_info(self, path: str) -> Tuple[str, str, str]:
        bucket_name = self.config[5:].split("/")[0]
        base_prefix = self.config[5 + len(bucket_name) + 1:]
        prefix = path[2:]
        if len(base_prefix) > 0:
            prefix = posixpath.join(base_prefix, prefix)
        return bucket_name, base_prefix, prefix

    def search(self, file_pattern: str) -> List[str]:
        """
        Gets paths for all files matching a glob pattern.

        :param file_pattern: (str) glob pattern of the file being searched for
        :return: list of strings which are paths to the found files
        """
        bucket_name, base_prefix, prefix = self._extract_info(self.path)
        bucket = self._engine.resource.Bucket(bucket_name)
        if len(prefix) > 0 and not prefix.endswith("/"):
            prefix = prefix + "/"
        all_objects = bucket.objects.filter(Prefix=prefix)
        if len(prefix) > 0:
            file_pattern = posixpath.join(prefix, file_pattern)
        all_files = [obj.key for obj in all_objects]
        return [posixpath.join("./", posixpath.relpath(self._clean_name(file), base_prefix)) for file in all_files if
                self._glob_match(file_pattern, self._clean_name(file))]

    def copy_file(self, new_path: FilePath) -> None:
        """
        Copies self.path file to a new file path.

        :param new_path: (FilePath) destination path for file to be copied to
        :return: None
        """
        old_bucket_name, old_file_name, _ = self._engine._split_s3_path(self.path)
        new_bucket_name, new_file_name, _ = self._engine._split_s3_path(new_path)

        self._engine.resource.Object(new_bucket_name, new_file_name).copy_from(
            CopySource=f"{old_bucket_name}/{old_file_name}")

    def copy_folder(self, new_folder: str) -> None:
        """
        Copies one folder to another folder.

        :param new_folder: (str) the directory to where the folder will be copied to
        :return: None
        """
        old_bucket_name, old_folder_name, _ = self._engine._split_s3_path(self.path)
        new_bucket_name, new_folder_name, _ = self._engine._split_s3_path(new_folder)

        old_bucket = self._engine.resource.Bucket(old_bucket_name)
        new_bucket = self._engine.resource.Bucket(new_bucket_name)

        for obj in old_bucket.objects.filter(Prefix=f"{old_folder_name}/"):
            old_source = {'Bucket': old_bucket_name,
                          'Key': obj.key}

            new_key = obj.key.replace(old_folder_name, new_folder_name, 1)
            new_obj = new_bucket.Object(new_key)
            new_obj.copy(old_source)

    @staticmethod
    def _split_s3_path(storage_path: Union[FilePath, str]) -> Tuple[str, str, str]:
        """
        Splits the path into bucket, file name, and short_file_name.

        :param storage_path: (Union[FilePath, str]) path to be split
        :return: (Tuple[str, str, str]) bucket_name, file_name, short_file_name
        """
        path = storage_path.replace("s3://", "")
        path = path.split("/")
        bucket_name = path[0]
        file_name = "/".join(path[1:])
        short_file_name = path[-1]
        return bucket_name, file_name, short_file_name

    @staticmethod
    def _clean_name(text: str) -> str:
        """
        Cleans the string from quotes.

        :param text: (str) string to be cleaned
        :return: (str) cleaned string
        """
        res = unquote(text)
        return res

    @staticmethod
    def _glob_match(pattern: str, file: str):
        return bool(monolith_globre_match(pattern, file))

    def rename_file(self, new_name: str) -> None:
        """
        Checks file to rename exists, checks new name not already taken.
        Copies file object to new location/name within bucket and then deletes old object.

        :param new_name: (str) new file name without extension
        :return: None
        :raises: (S3ProcessesAdapterError) if file does not exist or new name is taken
        """
        ext = f".{self.path.split('.')[-1]}"
        new_path = FilePath("/".join(self.path.split("/")[:-1]) + f"/{new_name}" + ext)
        parent_dir = FilePath("/".join(self.path.split("/")[:-1]) + "/")

        files, dirs = self.ls(path=parent_dir)
        if not self.exists():
            raise S3ProcessesAdapterError("File does not exist to rename")
        if self.check_name_taken(name=new_name, existing_names=[files for files in files.keys()] + dirs):
            raise S3ProcessesAdapterError("New file name already taken by a file or directory in this folder")

        self.copy_file(new_path=new_path)
        self.delete_file()

    def rename_folder(self, new_name: str) -> None:
        """
        Checks new name not already taken.
        Copies folder object and contents to new location/name within bucket and then deletes old object and contents.

        :param new_name: (str) new folder name
        :return: None
        :raises: (S3ProcessesAdapterError) if folder does not exist or new name is taken
        """
        new_path = FilePath("/".join(self.path.split("/")[:-1]) + f"/{new_name}")

        if self.exists(path=new_path):
            raise S3ProcessesAdapterError("New folder name already taken by a file or directory in this folder")

        self.copy_folder(new_folder=new_path)

        self.delete_folder()

    def move_file(self, destination_folder: str) -> None:
        """
        Checks new file path not already taken.
        Copies file object to new location/name within bucket and then deletes old object.

        :param destination_folder: (str) folder file is to be moved to
        :return: None
        :raises: (S3ProcessesAdapterError) if destination filepath already exists
        """
        file = self.path.split("/")[-1]
        new_path = FilePath(f"{destination_folder}/{file}")
        if self.exists(path=new_path):
            S3ProcessesAdapterError("File with same name already exists in destination folder")

        self.copy_file(new_path=new_path)
        self.delete_file()

    def move_folder(self, destination_folder: str) -> None:
        """
        Checks new folder path not already taken.
        Copies folder object and contents to new location/name within bucket and then deletes old folder.

        :param destination_folder: (str) folder to be moved to
        :return: None
        :raises: (S3ProcessesAdapterError) if destination filepath already exists
        """
        folder = self.path.split("/")[-1]
        new_path = FilePath(f"{destination_folder}/{folder}")

        if self.exists(path=new_path):
            S3ProcessesAdapterError("File/Folder with same name already exists in folder")

        self.copy_folder(new_folder=new_path)

        self.delete_folder()

    def batch_move(self, paths: List[str], destination_folder: str) -> None:
        """
        Iterates through list of paths within the self.path directory, ascertains if file or folder, then calls
        appropriate move method.

        :param paths: (List[str]) file and folder paths to be moved
        :param destination_folder: (str) folder path to be moved to
        :return: None
        """
        origin_folder = self.path
        for path in paths:
            dest_path = FilePath(f"{destination_folder}/{path}")
            if dest_path.get_file_type(dest_path.to_string()) is not None:
                self.path = FilePath(f"{self.path}/{path}")
                self.move_file(destination_folder=destination_folder)
            else:
                self.path = FilePath(f"{self.path}/{path}")
                self.move_folder(destination_folder=destination_folder)
            self.path = origin_folder

    @staticmethod
    def check_name_taken(name: str, existing_names: List[str]) -> bool:
        """
        Checks if a name string is in a list of strings.
        :param name: (str) file or folder name
        :param existing_names: (List[str]) list of strings to check
        """
        if name in existing_names:
            return True
        else:
            return False

    def _strip_path_slash(self) -> None:
        """
        self.path has any trailing '/' stripped to account for root folder level paths e.g. s3://example-bucket/

        :return: None
        """
        self.path = FilePath(self.path.rstrip("/"))





class BucketManagetError(Exception):

    def __init__(self, message):
        super().__init__(message)


class FileManagerError(Exception):

    def __init__(self, message):
        super().__init__(message)


class V1EngineError(Exception):

    def __init__(self, message):
        super().__init__(message)



from unittest import TestCase, main

from monolith_filemanager.components.path_map import PathMap


class TestPathMap(TestCase):

    def test___init__(self):
        input_dict = {"one": "1", "two": "2", "three": "3"}
        test = PathMap(file_map=input_dict)
        self.assertEqual(input_dict, test)

    def test_map_path(self):
        test = PathMap(file_map={"one": "s3://1", "two": "2", "three": "3"})

        self.assertEqual("s3://1/to/path.txt", test.map_path(path="one/to/path.txt"))
        self.assertEqual("2/to/path.txt", test.map_path(path="two/to/path.txt"))
        self.assertEqual("3/to/path.txt", test.map_path(path="three/to/path.txt"))
        self.assertEqual("four/to/path.txt", test.map_path(path="four/to/path.txt"))


if __name__ == "__main__":
    main()



from unittest import TestCase, main
from unittest.mock import patch, MagicMock

from monolith_filemanager.file.standard_pickle_file import StandardPickleFile


class TestStandardPickleFile(TestCase):

    @patch("monolith_filemanager.file.standard_pickle_file.File.__init__")
    def test___init__(self, mock_file):
        StandardPickleFile(path="test")
        mock_file.assert_called_once_with(path="test")

    @patch("monolith_filemanager.file.standard_pickle_file.pickle")
    @patch("monolith_filemanager.file.standard_pickle_file.open")
    @patch("monolith_filemanager.file.standard_pickle_file.StandardPickleFile.__init__")
    def test_read(self, mock_init, mock_open, mock_pickle):
        mock_init.return_value = None
        test = StandardPickleFile(path="test")
        test.path = "test"
        out_come = test.read()

        mock_open.return_value.__enter__.assert_called_once_with()
        self.assertEqual(mock_pickle.load.return_value, out_come)
        mock_open.assert_called_once_with("test", "rb")
        mock_pickle.load.assert_called_once_with(mock_open.return_value.__enter__.return_value)

    @patch("monolith_filemanager.file.standard_pickle_file.pickle")
    @patch("monolith_filemanager.file.standard_pickle_file.open")
    @patch("monolith_filemanager.file.standard_pickle_file.StandardPickleFile.__init__")
    def test_write(self, mock_init, mock_open, mock_pickle):
        mock_init.return_value = None
        test = StandardPickleFile(path="test")
        test.path = "test"
        mock_data = MagicMock()
        test.write(data=mock_data)

        mock_open.assert_called_once_with("test", "wb")
        mock_pickle.dump.assert_called_once_with(mock_data, mock_open.return_value.__enter__.return_value)


if __name__ == "__main__":
    main()



from typing import Optional, Dict, Any

from .errors import BucketManagetError


class BucketManager:
    """
    This is a class managing the operations for s3 buckets.

    Attributes:
        resource (Any): object managing the resource for s3 connection (None unless overridden by subclass like V1Engine)
        client (Any): object managing the client for s3 connection (None unless overridden by subclass like V1Engine)
        buckets (dict): keeps reference of buckets. Keys are made of bucket names
        bucket_contents (dict) keeps reference of bucket contents. Keys are made of bucket names
    """
    def __init__(self):
        """
        The constructor for the BucketManager class.
        """
        self.resource: Optional[Any] = None
        self.client: Optional[Any] = None
        self.buckets: Dict = dict()
        self.bucket_contents: Dict = dict()

    def get_buckets(self) -> None:
        """
        Fills self.buckets with all the buckets from S3.

        :return: None
        """
        self.buckets = dict()
        for bucket in self.resource.buckets.all():
            self.buckets[str(bucket.meta.data["Name"])] = bucket

    def get_bucket(self, name: str):
        """
        Gets bucket object from self.buckets.

        :param name: (str) name of bucket required
        :return: bucket object
        """
        self._check_bucket_cache()
        bucket = self.buckets.get(name)
        if not bucket:
            raise BucketManagetError(message="bucket {} not available. please pick from: {}".format(name,
                                                                                                    self.buckets.keys()
                                                                                                    ))
        return bucket

    def get_bucket_contents(self, bucket_name: str) -> None:
        """
        Gets a list of files from a bucket and stores it in self.bucket_contents with the key of the bucket name.
        
        :param bucket_name: (str) name of the bucket required
        :return: None
        """
        self.get_bucket(name=bucket_name)
        package = dict()
        contents = self.client.list_objects_v2(Bucket=bucket_name)["Contents"]
        package["keys"] = [i["Key"] for i in contents]
        package["contents"] = contents
        self.bucket_contents[bucket_name] = package

    def _check_bucket_cache(self) -> None:
        """
        Fills self.buckets if self.buckets is empty (private).
        
        :return: None
        """
        if len(self.buckets) == 0:
            self.get_buckets()

    def create_bucket(self, name: str) -> None:
        """
        Creates a bucket in S3 storage.
        
        :param name: (str) name of bucket you want to create.
        :return: None
        """
        self._check_bucket_cache()
        if self.buckets.get(name):
            raise BucketManagetError(message="{} already exists".format(name))
        # TODO look into defining the region for the create_bucket function
        self.resource.create_bucket(Bucket=name)
        self.get_buckets()

    def delete_bucket(self, name: str, force: Optional[bool] = False) -> None:
        """
        Deletes bucket (does not support buckets that have contents).
        
        :param name: (str) name of bucket you want to delete
        :param force: (bool) if True, will delete even if bucket has contents
        :return: None
        """
        self._check_bucket_cache()
        if not self.buckets.get(name):
            raise BucketManagetError(message="{} does not exist".format(name))
        self.resource.delete_bucket(Bucket=name, DeleteIfNonEmpty=force)
        del self.buckets[name]



import setuptools
from setuptools import find_packages
from setuptools.command.build_py import build_py as build_py_orig


extras_packages = {
    "flask": ["flask", "tensorflow", "boto3", "protobuf", "jinja2", "itsdangerous", "werkzeug",
              "markupsafe"],
    "3d": ["pyvista", "cqkit", "cadquery"],
    "matlab": ["scipy"]
}

__version__ = '3.0.0'

class CustomBuildPy(build_py_orig):
    """
    subclass build_py so that we collect no .py files inside the built pip package
    this is done by overriding build_packages method with a noop
    """
    def build_packages(self):
        pass


with open("README.md", "r") as fh:
    long_description = fh.read()

directives = {
    'language_level': 3,
    'always_allow_keywords': True
}


setuptools.setup(
    name="monolith_filemanager",
    version=__version__,
    author="Monolith AI",
    description="Python package for reading and writing files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MonolithAILtd/monolith-filemanager",
    install_requires=[
        'h5py==2.10.0',
        'joblib==0.15.0',
        'flask==1.1.4',
        'jinja2==2.11.3',
        'itsdangerous==1.1.0',
        'werkzeug==1.0.1',
        'MarkupSafe==2.0.1',
        'tensorflow==2.2.0',
        'protobuf==3.19.6',
        'numpy==1.22.4',
        'pandas==1.0.5',
        'dask[complete]==2020.12.0',
        'distributed==2021.1.1',
        'dill==0.2.9',
        'pyyaml==5.1.2',
        'boto3==1.10.5',
        'botocore==1.13.5',
        'requests==2.22.0',
        'openpyxl==3.0.7',
        'pyarrow==0.16.0',
        'xlwt==1.3.0',
        'xlrd==1.2.0',
        'scipy==1.4.1',
        'cqkit==0.5.1',
        'cadquery==2.2.0',
        's3fs==0.3.0',
    ],
    extras_require={
        'flask': extras_packages["flask"],
        '3d': extras_packages["3d"],
        'matlab': extras_packages["matlab"]
    },
    packages=find_packages(exclude=("tests",)),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    tests_require=['pytest'],
    entry_points={
        'console_scripts': [
            'file-hello = monolith_filemanager.console_commands.hello:print_logo',
            'file-install-flask = monolith_filemanager.console_commands.install_flask:install_flask',
            'file-install-tensorflow = monolith_filemanager.console_commands.install_tensorflow:install_tensorflow',
            'file-install-aws = monolith_filemanager.console_commands.install_boto:install_boto'
        ],
    }
)



import subprocess
import sys


def install_tensorflow():
    """
    Installs the tensorflow requirement for the package.

    Returns: None
    """
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'tensorflow>=2.1.0'])



import os
from typing import Any, Union

import cqkit
from cadquery import cq

from .base import File
from .errors import StlObjectFileError
from ..path import FilePath


class StlObjectFile(File):
    """
    This is a class for managing the reading and writing of stl objects.
    """
    SUPPORTED_FORMATS = ['stp', 'step', 'igs', 'iges']

    def __init__(self, path: Union[str, FilePath]) -> None:
        """
        The constructor for the CustomDataFile class.

        :param path: (str/FilePath) path to the file
        """
        super().__init__(path=path)

    def read(self, **kwargs) -> cq.Workplane:
        """
        Gets data from file defined by the file path.
        """

        _, file_ext = os.path.splitext(self.path)
        workplane = None

        if file_ext in [".stp", ".step"]:
            # The tag can be used to identify the workplane later e.g. in error reports
            workplane = cqkit.import_step_file(self.path).tag(self.path)
        elif file_ext in [".igs", ".iges"]:
            workplane = cqkit.import_iges_file(self.path).tag(self.path)

        # This logic will only ever be exercised if the SUPPORTED_FORMATS array above is updated
        # out-of-step with any CQKit methods used in this method
        if not workplane:
            raise StlObjectFileError(f"Unsupported file type: {file_ext}")

        return workplane

    def write(self, data: Any, **kwargs) -> None:
        """
        Writes data to file.

        :param data: (python object) data to be written to file
        :return: None
        """

        data.save(self.path)



import globre
from parameterized import parameterized
from unittest import TestCase

from monolith_filemanager.adapters.utils import monolith_globre_match

class TestUtils(TestCase):

    @parameterized.expand([
        ["*",             "/path/to/foo"                 ],
        ["**",            "/path/to/foo"                 ],
        ["/path/*",       "/path/to/foo"                 ],
        ["/path/**",      "/path/to/foo"                 ],
        ["/path/to/*",    "/path/to/foo"                 ],
        ["/path?to",      "/path/to"                     ],
        ["/path[!abc]to", "path/to"                      ],
        ["a/b/*/f.txt",   "a/b/c/f.txt"                  ],
        ["a/b/*/f.txt",   "a/b/q/f.txt"                  ],
        ["a/b/*/f.txt",   "a/b/c/d/f.txt"                ],
        ["a/b/*/f.txt",   "a/b/c/d/e/f.txt"              ],
        ["/foo/bar/*",    "/foo/bar/baz"                 ],
        ["/foo/bar/*",    "/spam/eggs/baz"               ],
        ["/foo/bar/*",    "/foo/bar/bar"                 ],
        ["/*/bar/b*",     "/foo/bar/baz"                 ],
        ["/*/bar/b*",     "/foo/bar/bar"                 ],
        ["/*/[be]*/b*",   "/foo/bar/baz"                 ],
        ["/*/[be]*/b*",   "/foo/bar/bar"                 ],
        ["/foo*/bar",     "/foolicious/spamfantastic/bar"],
        ["/foo*/bar",     "/foolicious/bar"              ]
    ])
    def test_monolith_globre_match(self, shell_pattern: str, string: str):
        # Test that our home solution is equivalent to globre
        self.assertEqual(bool(monolith_globre_match(shell_pattern, string)),
                         bool(globre.match(shell_pattern, string)))



from typing import Union, Optional

import dask.dataframe as dd
import pandas as pd
from dask import delayed
from dask.callbacks import Callback

from .base import File
from .errors import PandasFileError
from ..path import FilePath

PandasLoadMethod = Union[pd.read_parquet, pd.read_csv, pd.read_excel, pd.read_table, pd.read_table]
DataFrameType = Union[pd.DataFrame, dd.DataFrame]


def dask_read_excel(path: str, **kwargs) -> dd.DataFrame:
    delayed_df = delayed(pd.read_excel)(path)
    return dd.from_delayed(delayed_df)


class PandasFile(File):
    """
    This is a class for managing the reading and writing of data to files for pandas data frames.
    """
    PANDAS_LOADING_METHODS = {
        "parquet": pd.read_parquet,
        "csv": pd.read_csv,
        "xls": pd.read_excel,
        "xlsx": pd.read_excel,
        "dat": pd.read_table,
        "data": pd.read_table
    }

    DASK_LOADING_METHODS = {
        "parquet": dd.read_parquet,
        "csv": dd.read_csv,
        "xls": dask_read_excel,
        "xlsx": dask_read_excel,
        "dat": dd.read_table,
        "data": dd.read_table
    }

    PANDAS_SUPPORTED_FORMATS = list(PANDAS_LOADING_METHODS)
    DASK_SUPPORTED_FORMATS = list(DASK_LOADING_METHODS)
    SUPPORTED_FORMATS = list(set(PANDAS_SUPPORTED_FORMATS + DASK_SUPPORTED_FORMATS))

    def __init__(self, path: Union[str, FilePath]) -> None:
        """
        The constructor for the PandasFile class.

        :param path: (str/FilePath) path to the file
        """
        super().__init__(path=path)

    def read(self, lazy: bool = False, chunk_size: Union[int, str] = '256MB', reset_index_if_eager: bool = True,
             **kwargs) -> DataFrameType:
        """
        Gets data from file defined by file path.
        If `lazy` is True, then returns a Dask DataFrame.
        If `lazy` is False, then returns a Pandas DataFrame. The index can be reset to avoid inconsistency with how Dask
            stores DataFrame indexes.

        :param lazy: (bool) Whether reading should be lazy (returns dask DataFrame) or eager (returns pandas DataFrame)
        :param chunk_size: (dask compatible int or str) size in bytes of chunks to read. Only used if lazy == True
        :param reset_index_if_eager: (bool) Whether or not to reset the index of the read dataframe.
            Only resets the index of pandas dataframes
        :return: (DataFrameType) Data from file
        """
        storage_options = {'config_kwargs': {'max_pool_connections': 32}, 'skip_instance_cache': True}
        df = self._read_dask(chunk_size, storage_options=storage_options, **kwargs) if lazy \
            else self._read_dask(chunk_size, storage_options=storage_options, **kwargs).compute()

        if reset_index_if_eager and not lazy:
            # Reset the index of pandas dataframes if requested
            df = df.reset_index(drop=True)

        return df

    def write(self, data: DataFrameType, repartition: bool = False, divisions: Union[int, str] = '64MB',
              scheduler: str = 'threads', cb: Optional[Callback] = None, **kwargs) -> None:
        """
        Writes data to file.

        :param data: (pandas or dask data frame) data to be written to file
        :param repartition: (bool) whether or not to repartition the dataframe to a given chunk size. Default to False.
        :param divisions: (list of int, or int or str) dask-compatible maximum partition size.
                                    Interpreted as index of partitions if a list of ints
                                    Interpreted as number of bytes if str,
                                    Interpreted as number of divisions if int.
        :param scheduler: (str) Dask local scheduler type to use for computation.
            Choose from "threads", "single-threaded", or "processes"
        :param cb: (optional dask Callback) A dask-compatible callback for updates during computation of the dask graph.
        :return: None
        """

        if not isinstance(data, (pd.DataFrame, dd.DataFrame)):
            raise PandasFileError(message=f'Data passed to write method isn\'t a pandas or dask DataFrame. '
                                          f'Please use a DataFrame for {self.DASK_SUPPORTED_FORMATS}')

        if isinstance(data, pd.DataFrame):
            data = dd.from_pandas(data, npartitions=1, sort=False)

        if repartition:
            data = data.repartition(divisions=divisions)

        if cb is None:
            self._write_functions(data, self.path, scheduler)
        else:
            with cb:
                self._write_functions(data, self.path, scheduler)

    @staticmethod
    def _write_functions(df: DataFrameType, path: FilePath, scheduler: str) -> None:
        """
        Function for mapping file type to correct write method.

        :param df: (pandas or dask data frame) data to be written to file
        :param path: (FilePath) the path to save the dataframe to.
        :param scheduler: (str) Dask local scheduler type to use for computation.
            Choose from "threads", "single-threaded", or "processes"
        """
        if path.file_type in ('xls', 'xlsx'):
            df.compute(scheduler=scheduler).to_excel(path),
        elif path.file_type in ('csv', 'dat', 'data'):
            df.to_csv(path, compute_kwargs={'scheduler': scheduler}, single_file=True)
        elif path.file_type in ('parquet',):
            df.to_parquet(path, compute_kwargs={'scheduler': scheduler})
        else:
            raise ValueError(f'Failed writing dataframe to file. File type {path.file_type} not known.')

    def _read_dask(self, chunk_size: Union[int, str], **kwargs) -> dd.DataFrame:
        """
        Read from file using dask loading method

        :param chunk_size: (int or str) dask-compatible maximum partition size. Interpreted as number of bytes.
        :return: dask DataFrame from file
        """
        if self.path.file_type not in self.DASK_SUPPORTED_FORMATS:
            raise PandasFileError(f'File type {self.path.file_type} not supported for lazy loading.')

        if self.path.file_type in ('xls', 'xlsx'):
            return self.DASK_LOADING_METHODS[self.path.file_type](self.path, **kwargs)
        elif self.path.file_type in ('parquet',):
            return self.DASK_LOADING_METHODS[self.path.file_type](self.path, chunksize=chunk_size, **kwargs)

        return self.DASK_LOADING_METHODS[self.path.file_type](self.path, blocksize=chunk_size, **kwargs)

    @staticmethod
    def supports_s3():
        return True
